/*
  Program:      Mine Mayhem
  File: 	fieldhxr.cc
  Date: 	26 October to 3 December, 1997, version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Everything to do with the hex-tiled rectangular minefield.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.

  v2.10: removed PanLevel() function, use pan_wid variable instead.
*/

#include "minefld.h"
#include "options.h"
#include "tiles.h"
#include "board.h"


RectHexField::RectHexField() : SquareField(false)
{
  MaxWidth	 = (max_width / HEXAGONW) - 1;
  MaxHeight	 =
  rows_allocated = (max_height - HEXAGOND) / HEXAGONR;

  olddim	 = Coord(MaxWidth, MaxHeight);

  CreateMinefield(MaxWidth);
}


/* ------------------------------   canWrap   -------------------------------

  Only the even-height fields can wrap.

*/

bool RectHexField::canWrap()
{
 return (layout.dim.y % 2 == 0);
}


/* ------------------------------   Convert   -------------------------------

  Convert the co-ordinate to a tile.
  Returns false if the co-ordinate is not on the grid, otherwise true.
  false is also returned if the hexagon is not next to an open hexagon in the
  cross game.

*/

bool RectHexField::Convert(int x, int y, Tile*& t)
{
  Coord b = GetBoardDimensions(layout.dim);
  y -= HEXAGOND/2;
  if (x < 0 || x >= b.x ||
      y < 0 || y >= b.y) return false;

  y /= HEXAGONR;
  if (y >= layout.dim.y) return false;

  if (y % 2) x -= (HEXAGONW/2);
  if (x < 0) return false;
  x /= HEXAGONW;
  if (x >= layout.dim.x) return false;

  t = &grid[y][x];
  return CrossValid(t);
}


/* -----------------------------   EveryTile   ------------------------------

  Create the list of hexagons on the minefield and adjust the edges'
  surrounding hexagons (if create is false).

  If create is true then initialise all the display co-ordinates and
  surrounding tiles for each tile.

*/

void RectHexField::EveryTile(bool create)
{
  if (!create)
  {
    SquareField::EveryTile();
    pan_wid = layout.dim.x * HEXAGONW - HEXAGONW/2;
  }
  else
  {
    layout.dim = Coord(MaxWidth, MaxHeight);
    for (int y = 0; y < MaxHeight; y++)
    {
      for (int x = 0; x < MaxWidth; x++)
      {
	grid[y][x].loc.x    = x*HEXAGONW + HEXAGONW/2 * (y%2);
	grid[y][x].loc.y    = y*HEXAGONR;
	grid[y][x].surround = new Tile*[6];
	Surround(x, y);
      }
    }
  }
}


/* ------------------------------   Surround   ------------------------------

  Determine the surrounding hexagons, where the order is:

   5 3
  4 - 1       The - is (x, y)
   2 0

  The reason for this is to help minimise opening hexagons in the cross game.

  In the array, it looks like:	Odd lines     Even lines
				  5 3		  5 3
				  4 - 1 	4 - 1
				  2 0		  2 0

*/

void RectHexField::Surround(int x, int y)
{
  tile = &grid[y][x];
  tile->count = 0;
  bool temp = isWrap;
  if (isWrap) isWrap = canWrap();
  Validate(x + y%2,	y+1);
  Validate(x + 1,	y);
  Validate(x - (y+1)%2, y+1);
  Validate(x + y%2,	y-1);
  Validate(x - 1,	y);
  Validate(x - (y+1)%2, y-1);
  isWrap = temp;
}


/* ---------------------------	 AdjustSurround   ---------------------------

  When the minefield's dimensions change, the edges need to be re-surrounded.
  Redo the old layout's edges, then the new ones. If it's wrap, then all edges
  are required to be surrounded.

*/

void RectHexField::AdjustSurround()
{
  bool temp = isWrap;
  if (isWrap) isWrap = canWrap();
  SquareField::AdjustSurround();
  isWrap = temp;
}


/* ----------------------------   DisplayTile	-----------------------------

  Display the hexagon t using which bitmap.

*/

void RectHexField::DisplayTile(Tile* t, int which, bool refresh)
{
  Board->DisplayTile(t->loc.x, t->loc.y, HEXAGONW-1, HEXAGONH-1,
		     Hexagon[which], refresh);
}


/* -------------------------   GetBoardDimensions   -------------------------

  Given the dimensions of the minefield, return the dimensions of the board.

*/

Coord RectHexField::GetBoardDimensions(cCoord& dim)
{
  return Coord(dim.x * HEXAGONW + HEXAGONW/2, dim.y * HEXAGONR + HEXAGOND);
}


/* ------------------------------   Outline   -------------------------------

  Draw the outline around the minefield.

*/

void RectHexField::Outline(ViewBuffer* buf, int x1, int y1, int x2, int y2)
{
  int lx = x1, rx = x2 - HEXAGONW,
      ty = y1, by = y1;

  // Left side.
  for (int y = 0; y < layout.dim.y/2; ++y)
  {
    BitBltTransparent(buf, lx, ty, Hexagon[7], HEXAGONW/2-1, 0,
					       HEXAGONW-1,   HEXAGONR);
    BitBltTransparent(buf, lx, ty+HEXAGONR+1, Hexagon[8],
		      HEXAGONW/2-1, HEXAGONR-HEXAGOND, HEXAGONW-1, HEXAGONR);
    BitBltTransparent(buf, lx+HEXAGONW/2, ty+HEXAGONH+1, Hexagon[8],
		      HEXAGONW/2-1, 0, HEXAGONW/2-1, HEXAGONR-1);
    ty += HEXAGONR*2;
  }
  if (layout.dim.y % 2 == 1)
  {
    BitBltTransparent(buf, lx, ty, Hexagon[7], HEXAGONW/2-1, 0,
					       HEXAGONW-1,   HEXAGONR);
    ty += HEXAGONR;
  }
  else lx += HEXAGONW/2;
  BitBltTransparent(buf, lx, ty+1, Hexagon[8],
		    HEXAGONW/2-1, HEXAGONR-HEXAGOND, HEXAGONW-1, HEXAGONR);

  // Right side.
  BitBltTransparent(buf, rx, by, Hexagon[7], 0, 0, HEXAGONW/2, HEXAGONR);
  rx += HEXAGONW/2; by += HEXAGONR;
  for (int y = 0; y < (layout.dim.y-1)/2; ++y)
  {
    BitBltTransparent(buf, rx, by+HEXAGOND+1, Hexagon[8], 0, 0,
							  HEXAGONW/2, HEXAGONR);
    BitBltTransparent(buf, rx, by+HEXAGONH+1, Hexagon[8], HEXAGONW/2, 0,
							 HEXAGONW/2, HEXAGONR);

    by += HEXAGONR*2;
  }
  if (layout.dim.y % 2 == 0)
  {
    BitBltTransparent(buf, rx, by+HEXAGOND+1, Hexagon[8], 0, 0,
							  HEXAGONW/2, HEXAGONR);
  }
  else
  {
    BitBltTransparent(buf, rx-HEXAGONW/2, by+1, Hexagon[8],
			   0, HEXAGONR-HEXAGOND, HEXAGONW/2, HEXAGONR);
  }

  lx = x1+1 + HEXAGONW/2; rx = lx + (layout.dim.y % 2 == 0)*HEXAGONW/2;
  ty = y1;		  by = y2 - HEXAGOND;
  for (int x = 1; x < layout.dim.x; ++x)
  {
    // Top
    BitBltTransparent(buf, lx, ty, Hexagon[7], 0, 0, HEXAGONW-1, HEXAGOND);
    // Bottom
    BitBltTransparent(buf, rx, by, Hexagon[8], 0, HEXAGONR-HEXAGOND,
					       HEXAGONW-1, HEXAGONR);
    lx += HEXAGONW;
    rx += HEXAGONW;
  }
}


/* -------------------------   ReadIni & WriteIni   -------------------------

  As you can see, it uses the "[ Hexagons ]" section and times are kept in
  the "minehxr.tym" file.

*/

void RectHexField::ReadIni(IniFile& ini)
{
  Minefield::ReadIni(ini, "Hexagons", "minehxr.tym");
}

void RectHexField::WriteIni(IniFile& ini)
{
  Minefield::WriteIni(ini, "Hexagons", "minehxr.tym");
}
