/*
  Program:      Mine Mayhem
  File: 	fieldsqr.cc
  Date: 	July, 1996 to February,   1997, version 1.00.
		26 October to 3 December,	version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Everything to do with the rectangular minefield.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.

  v2.00: lots of changes due to the board -> minefield conversion.
  v2.10: removed PanLevel() function, use pan_wid variable instead.
*/

#include "minefld.h"
#include "options.h"
#include "tiles.h"
#include "board.h"


SquareField::SquareField(bool create) : Minefield()
{
  Beginner	 = Layout( 8,  8, 10);
  Intermediate	 = Layout(16, 16, 40);
  Expert	 = Layout(30, 16, 99);

  MinWidth	 =
  MinHeight	 = 8;
  MaxWidth	 = max_width  / SQUAREW;
  MaxHeight	 =
  rows_allocated = max_height / SQUAREH;

  olddim	 = Coord(MaxWidth, MaxHeight);

  if (create) CreateMinefield(MaxWidth);
}


/* --------------------------------   Area   --------------------------------

  Area of a rectangle - anyone not know this?

*/

int SquareField::area(cCoord& dim)
{
  return (dim.x * dim.y);
}


/* -------------------------   CrossStart and End   -------------------------

  Return the beginning and finishing positions for the cross game.

*/

Tile* SquareField::CrossStart()
{
  return tiles[0];
}

Tile* SquareField::CrossEnd()
{
  return tiles[tile_count-1];
}


/* ----------------------------   RandomLayout	 ----------------------------

  Calculate the dimensions to satisfy the difficulty. diff is assumed to be
  valid (ie. between 2 and 10). There will always exist a layout to satisfy
  the difficulty; a dimension will be a multiple of it.

*/

void SquareField::RandomLayout(int diff)
{
  int a;
  do
  {
    Random.dim.x = random() % (MaxWidth  - MinWidth  + 1) + MinWidth;
    Random.dim.y = random() % (MaxHeight - MinHeight + 1) + MinHeight;
    a = area(Random.dim);
  }
  while (a % diff != 0);
  Random.mines = a / diff;
}


/* -----------------------------   AdjustGame	-----------------------------

  When the game changes from CLEAR to CROSS, need to ensure that the top and
  bottom corners are free from mines.

*/

void SquareField::AdjustGame()
{
  bool moving = true;
  while (moving)
  {
    moving = (MoveMine(tiles[0])			   ||
	      MoveMine(tiles[1])			   ||
	      MoveMine(tiles[layout.dim.x])		   ||
	      MoveMine(tiles[layout.dim.x+1])		   ||
	      MoveMine(tiles[tile_count-2 - layout.dim.x]) ||
	      MoveMine(tiles[tile_count-1 - layout.dim.x]) ||
	      MoveMine(tiles[tile_count-2])		   ||
	      MoveMine(tiles[tile_count-1]));
  }
}


/* -----------------------------   AdjustWrap	-----------------------------

  When the Wrap-around option is changed, the edges need to be recounted and
  re-surrounded (but not in that order...).

*/

void SquareField::AdjustWrap()
{
  int lx = layout.dim.x - 1,
      ly = layout.dim.y - 1;

  #define Adjust(t) if ((t)->value != MINE) (t)->value = Count((t), MINE)

  for (int x = 0; x <= lx; x++)
  {
    Surround(x, 0);   Surround(lx - x, ly);
    Adjust(tiles[x]); Adjust(tiles[tile_count-1 - x]);
  }
  for (int y = 1; y < ly; y++)
  {
    Surround(0, y);	 Surround(lx, y);
    Adjust(&grid[y][0]); Adjust(&grid[y][lx]);
  }

  #undef Adjust
}


/* ------------------------------   Convert   -------------------------------

  Convert the co-ordinate to a tile.
  Returns false if the co-ordinate is not on the grid, otherwise true.
  false is also returned if the square is not next to an open square in the
  cross game.

*/

bool SquareField::Convert(int x, int y, Tile*& t)
{
  if (x < 0 || x >= layout.dim.x*SQUAREW ||
      y < 0 || y >= layout.dim.y*SQUAREH) return false;
  x /= SQUAREW; y /= SQUAREH;

  t = &grid[y][x];
  return CrossValid(t);
}


/* -----------------------------   EveryTile   ------------------------------

  Create the list of squares on the minefield and adjust the edges' surrounding
  squares (if create is false).

  If create is true then initialise all the display co-ordinates and
  surrounding tiles for each tile.

*/

void SquareField::EveryTile(bool create)
{
  if (!create)
  {
    tile_count = area(layout.dim);
    int pos = 0;
    for (int y = 0; y < layout.dim.y; y++)
    {
      for (int x = 0; x < layout.dim.x; x++)
      {
	tiles[pos++] = &grid[y][x];
      }
    }
    AdjustSurround();
    pan_wid = (layout.dim.x - 1) * SQUAREW;
  }
  else
  {
    layout.dim = Coord(MaxWidth, MaxHeight);
    for (int y = 0; y < MaxHeight; y++)
    {
      for (int x = 0; x < MaxWidth; x++)
      {
	grid[y][x].loc.x    = x*SQUAREW;
	grid[y][x].loc.y    = y*SQUAREH;
	grid[y][x].surround = new Tile*[8];
	Surround(x, y);
      }
    }
  }
}


/* ------------------------------   Surround   ------------------------------

  Determine the surrounding squares, where the order is:

  7 5 3
  6 - 1        The - is (x, y)
  4 2 0

  The reason for this is to help minimise opening squares in the cross game.

*/

void SquareField::Validate(int c, int r)
{
  int w = layout.dim.x - 1,
      h = layout.dim.y - 1;

  if (Game == CROSS || !isWrap)
  {
    if (c < 0 || c > w || r < 0 || r > h) return;
  }
  else
  {
    c = (c < 0) ? w : (c <= w) ? c : 0;
    r = (r < 0) ? h : (r <= h) ? r : 0;
  }
  tile->surround[tile->count++] = &grid[r][c];
}

void SquareField::Surround(int x, int y)
{
  tile = &grid[y][x];
  tile->count = 0;
  Validate(x+1, y+1);
  Validate(x+1, y);
  Validate(x,	y+1);
  Validate(x+1, y-1);
  Validate(x-1, y+1);
  Validate(x,	y-1);
  Validate(x-1, y);
  Validate(x-1, y-1);
}


/* ---------------------------	 AdjustSurround   ---------------------------

  When the minefield's dimensions change, the edges need to be re-surrounded.
  Redo the old layout's edges, then the new ones. If it's wrap, then all edges
  are required to be surrounded.

*/

void SquareField::AdjustSurround()
{
  if (olddim == layout.dim) return;

  int x, y;
  Coord temp = layout.dim;

  layout.dim = Coord(MaxWidth, MaxHeight);
  for (y = 0; y < olddim.y; ++y)
  {
    Surround(olddim.x-1, y);
  }
  for (x = 0; x < olddim.x; ++x)
  {
    Surround(x, olddim.y-1);
  }

  layout.dim = temp;
  for (y = 0; y < layout.dim.y; ++y)
  {
    Surround(layout.dim.x-1, y);
  }
  for (x = 0; x < layout.dim.x-1; ++x)
  {
    Surround(x, layout.dim.y-1);
  }

  if (isWrap && Game != CROSS)
  {
    for (y = 0; y < layout.dim.y-1; ++y)
    {
      Surround(0, y);
    }
    for (x = 0; x < layout.dim.x-1; ++x)
    {
      Surround(x, 0);
    }
  }

  olddim = layout.dim;
}


/* ----------------------------   DisplayTile	-----------------------------

  Display the square t using which bitmap.

*/

void SquareField::DisplayTile(Tile* t, int which, bool refresh)
{
  Board->DisplayTile(t->loc.x, t->loc.y, SQUAREW-1, SQUAREH-1,
		     Square[which], refresh);
}


/* -------------------------   GetBoardDimensions   -------------------------

  Given the dimensions of the minefield, return the dimensions of the board.

*/

Coord SquareField::GetBoardDimensions(cCoord& dim)
{
  return Coord(dim.x * SQUAREW, dim.y * SQUAREH);
}


/* ------------------------------   Outline   -------------------------------

  Draw the outline around the minefield.

*/

void SquareField::Outline(ViewBuffer* buf, int x1, int y1, int x2, int y2)
{
  VLine(buf, x2, y1, y2, 149);		// 149 is grey
  HLine(buf, x1, x2, y2, 149);
  HLine(buf, x1, x2, y1, 235);		// 235 is white
  VLine(buf, x1, y1, y2, 235);
}


/* ------------------------------   Validate   ------------------------------

  Ensure the layout does not exceed the limits.

*/

void SquareField::Validate(Layout& layout)
{
  layout.dim.x = MinWidth  >? layout.dim.x <? MaxWidth;
  layout.dim.y = MinHeight >? layout.dim.y <? MaxHeight;
  int a        = area(layout.dim);
  layout.mines = (a/10 + (a%10 != 0)) >? layout.mines <? a/2;
}


/* -------------------------   ReadIni & WriteIni   -------------------------

  As you can see, it uses the "[ Squares ]" section and times are kept in
  the "minesqr.tym" file.

*/

void SquareField::ReadIni(IniFile& ini)
{
  Minefield::ReadIni(ini, "Squares", "minesqr.tym");
}

void SquareField::WriteIni(IniFile& ini)
{
  Minefield::WriteIni(ini, "Squares", "minesqr.tym");
}
