/*
  Program:      Mine Mayhem
  File: 	minefield.cc
  Date: 	July, 1996 to February,   1997, version 1.00.
		26 October to 3 December,	version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Everything to do with the minefield.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.

  v2.00: test AutoOpen in Clear(), rather than Open();
	 when highlighting too many marked, display flags as blanks;
	 open/mark/win/die sounds;
	 lots of changes due to the board -> minefield conversion.
  v2.10: modified panning, added SwapStereo option;
	 replaced DisplayMines() and DisplaySpaces() with DisplayProgess().
*/

#include "minefld.h"
#include "board.h"
#include "options.h"
#include "sound.h"
#include "faces.h"

#include "mainmenu.h"
extern MainMenuWindow* Main;


double Layout::diff() const
{
  return double(field->area(dim)) / mines;
}


/* -----------------------------   Minefield   ------------------------------

  Determine the largest board size.

*/

Minefield::Minefield() : isHiLite(false)
{
  //				     frame    margin outline
  max_width  = ws.GetDeskWidth() - (FWIDTH*2 + 4*2 +	2);

  //				    <----------- menu ----------->
  max_height = ws.GetDeskHeight() - (SysFont->height+FWIDTH*2+4+4) -
  //	       margin  time  margin  smiley    margin+outline	frame
	       (4+    DIGITH+  4 +  FACEICONH +    5 + 5     + FWIDTH*2);
}


/* --------------------------	CreateMinefield   ---------------------------

  Allocate memory for the minefield and pointers to it.

*/

void Minefield::CreateMinefield(int width)
{
  grid = new Tile*[rows_allocated];
  for (int j = 0; j < rows_allocated; j++)
    grid[j] = new Tile[width];

  tiles = new Tile*[width * rows_allocated];
  EveryTile(true);
}


/* -----------------------------   ~Minefield	-----------------------------

  Free the memory allocated.

*/

Minefield::~Minefield()
{
  for (int j = rows_allocated-1; j >= 0; j--)
    delete[] grid[j];
  delete[] grid;

  delete[] tiles;
}


/* ------------------------------   SetLevel   ------------------------------

  Set the layout according to the Level. If Level is zero then play the same
  level that was played before. If Level is invalid then play Beginner.
  If from_ini is true then the custom and random games are taken from the
  initialisation file, rather than selecting a new game. If a custom game is
  the same as a preset, the preset becomes the level; it's assumed a random
  game will never be the same as a preset.

*/

void Minefield::SetLevel(int Level, int from_ini)
{
  switch (Level)
  {
    case BEGINNER:     layout = Beginner;     break;
    case INTERMEDIATE: layout = Intermediate; break;
    case EXPERT:       layout = Expert;       break;
    case CUSTOM: if (from_ini || CustomDlg().Run())
                 {
                   layout = Custom;
                        if (layout == Beginner)     Level = BEGINNER;
                   else if (layout == Intermediate) Level = INTERMEDIATE;
                   else if (layout == Expert)       Level = EXPERT;
                 }
                 else Level = level;
                 break;
    default:
      if (Level < 2 || Level > 10)
      {
        Level = BEGINNER; layout = Beginner;
      }
      else
      {
	if (!from_ini) RandomLayout(Level);
        layout = Random;
      }
  }
  level = Level;
  EveryTile();
}


/* -------------------------------   NewGame   ------------------------------

  Start a new game with level Level.

*/

void Minefield::NewGame(int Level)
{
  StopAllSamples();

  if (Level != 0)
  {
    SetLevel(Level, false);
    Main->cmGame(true); 		// In case the field can no longer wrap
  }

  mines      =
  mynes      = layout.mines;
  spaces     = area(layout.dim) - mines;
  isStarted  =
  isFinished = false;

  Board->NewGame(GetBoardDimensions(layout.dim), mines, spaces);
  foralltiles
  {
    tiles[tile]->value	= 0;
    tiles[tile]->status = UNOPENED;
    DisplayTile(tiles[tile], UNOPENED, false);
  }
  Board->RefreshWindow();

  for (int j = 0; j < layout.mines; j++) PlaceMine();
  if (Game == CROSS) AdjustGame();
}


/* ---------------------------	 GameinProgress   ---------------------------

  Return true if a game is currently being played (including paused).

*/

bool Minefield::GameinProgress()
{
  return (isStarted && !isFinished);
}


/* -------------------------------   Count   --------------------------------

  Count around the tile t and return a value depending on what:
  MINE:     the number of mines;
  MARKED:   the number of marked tiles (not questions);
  UNOPENED: the number of unopened and marked tiles;
   1:	    increase the value of the surrounding tiles (returns 0);
  -1:	    decrease the value of the surrounding tiles (returns 0).

*/

int Minefield::Count(Tile* t, int what)
{
  int count = 0;
  around (t, loop)
  {
    switch (what)
    {
      case MINE:     if (t->surround[loop]->value  == MINE)   count++; break;
      case MARKED:   if (t->surround[loop]->status == MARKED) count++; break;
      case UNOPENED: if (t->surround[loop]->status == UNOPENED ||
			 t->surround[loop]->status == MARKED) count++; break;
      case -1: case 1:
	if (t->surround[loop]->value != MINE) t->surround[loop]->value += what;
    }
  }
  return count;
}


/* -----------------------------   PlaceMine   ------------------------------

  Pick a random location to place a mine on the field and increase the count
  around it.

*/

void Minefield::PlaceMine()
{
  int pos;

  do
  {
    pos = random() % tile_count;
  }
  while (tiles[pos]->value == MINE);

  tiles[pos]->value = MINE;
  Count(tiles[pos], 1);
}


/* ------------------------------   MoveMine   ------------------------------

  If there's a mine at t move it somewhere else, update the count and
  return true. If there's no mine, return false.
  The display is not updated if tiles have already been opened.
  There's also a very small chance that t could be rechosen.

*/

bool Minefield::MoveMine(Tile* t)
{
  if (t->value != MINE) return false;

  Count(t, -1);
  t->value = Count(t, MINE);
  PlaceMine();

  return true;
}


/* ----------------------------   NearestSpace	 ----------------------------

  Find the closest blank tile to t (without wrap), using displayed distance.
  Return false if there is no blank tile.

*/

bool Minefield::NearestSpace(Tile*& t)
{
  if (t->value == 0) return true;

  int blank = -1,
      x = t->loc.x, y = t->loc.y,
      dist,
      closest = max_width*max_width + max_height*max_height;

  foralltiles
  {
    if (tiles[tile]->value == 0)
    {
      dist = (tiles[tile]->loc.x - x) * (tiles[tile]->loc.x - x) +
	     (tiles[tile]->loc.y - y) * (tiles[tile]->loc.y - y);
      if (dist < closest)
      {
	closest = dist;
	blank	= tile;
      }
    }
  }
  if (blank == -1) return false;

  t = tiles[blank];
  return true;
}


/* -----------------------------   CrossValid	-----------------------------

  Return false if the tile is not next to an open tile in the cross game.

*/

bool Minefield::CrossValid(Tile* t)
{
  if (Game == CROSS && GameinProgress())
  {
    around (t, loop) if (t->surround[loop]->status == OPENED) return true;
    return false;
  }

  return true;
}


/* -----------------------------   LeftButton	-----------------------------

  Open or clear-around a tile. If the tile is blank, or automatic opening or
  clearing is on and the tile has all its mines marked, open all surrounding
  tiles. If it's the first press, start timing.

*/

void Minefield::LeftButton(int x, int y)
{
  if (isFinished) return;
  Tile* t;
  if (!Convert(x, y, t)) return;

  if (t->status != UNOPENED)
  {
    if (ClearButton == CLRLEFT) ClearAround(t);
    return;
  }

  if (!isStarted)
  {
    isStarted = true;
    if (Game == CROSS) t = CrossStart();
    else switch (Start)
    {
      case NO_REPOS:  break;
      case ONE_REPOS: MoveMine(t); break;
      case AUTO:      if (!NearestSpace(t)) MoveMine(t);
    }
    Board->StartTiming();
  }

  Open(t);
  if (!isFinished &&
      (t->value == 0 ||
       ((isAutoOpen || isAutoClear) && t->value == Count(t, MARKED))))
  {
    Clear(t);
  }
  else Board->DisplayProgress(mines, spaces, UPDATE_SPACES);
}


/* ----------------------------   RightButton	-----------------------------

  If the game is over, start a new game with the same dimensions.
  If automatic opening is on then unmarking is disabled.
  If automatic marking is on then marking is disabled altogether.
  Tiles can only be marked next to open tiles in the cross and auto-open games.
  The clear game allows any tile to be marked (should this change?).
  You cannot mark more tiles than there are mines (since this renders the
  all-marked finishing condition meaningless).

*/

void Minefield::RightButton(int x, int y)
{
  if (isFinished)
  {
    NewGame();
    return;
  }
  if (isAutoMark || (!isStarted && (Game == CROSS || isAutoOpen))) return;

  Tile* t;
  int temp = Game;
  if (isAutoOpen) Game = CROSS;         // Bit of a hack, but what the heck
  bool valid = Convert(x, y, t);
  Game = temp;
  if (!valid ||
      (isAutoOpen && t->status == MARKED) ||
      (mines == 0 && t->status == UNOPENED)) return;

  if (t->status == OPENED)
  {
    if (ClearButton == CLRRIGHT) ClearAround(t);
    return;
  }

  Mark(t);
}


/* ----------------------------   MiddleButton	 ----------------------------

  Only clears-around at the moment.

*/

void Minefield::MiddleButton(int x, int y)
{
  if (ClearButton == CLRMIDDLE && GameinProgress())
  {
    Tile* t;
    if (Convert(x, y, t)) ClearAround(t);
  }
}


/* --------------------------------   Open   --------------------------------

  Open the tile t and test for game over. If automatic marking is on then
  test all around the tiles that are around it.
  The tile is assumed to be unopened and unmarked (both mine and question).
  The space count display is not updated.

*/

void Minefield::Open(Tile* t)
{
  t->status = OPENED;
  if (t->value == MINE)
  {
    //t->value = DEADMINE;
    DisplayTile(t, DEADMINE);
    Board->DisplayProgress(--mines, spaces, UPDATE_MINES);
    GameOver(true); return;
  }
  else
  {
    if (isSound && OpenSnd) PlaySampleFVP(open_snd, open_snd->frequency,
					  open_snd->volume, PanLevel(t->loc));
    DisplayTile(t, t->value);
    --spaces;
    if (GameOver()) return;
  }

  if (isAutoMark)
  {
    // The centre tile isn't included in the surrounding tiles.
    if (t->value == Count(t, UNOPENED))
    {
      around (t, j)
      {
	if (t->surround[j]->status == UNOPENED)
	{
	  Mark(t->surround[j]); if (isFinished) return;
	}
      }
    }
    around (t, j)
    {
      if (t->surround[j]->status == OPENED &&
	  t->surround[j]->value  == Count(t->surround[j], UNOPENED))
      {
	around (t->surround[j], k)
        {
	  if (t->surround[j]->surround[k]->status == UNOPENED)
	  {
	    Mark(t->surround[j]->surround[k]); if (isFinished) return;
	  }
        }
      }
    }
  }
}


/* -------------------------------   Clear   --------------------------------

  Clear around the selected tile. This is used to automatically open tiles
  that have no mines around them, and to clear around a tile that has all
  its mines marked. For the clear game, a queue is used to keep track of
  tiles to be opened. This creates a spiral effect which looks better than a
  recursive call. Doing this for the cross game could open unnecessary tiles,
  so a straight recursive call is used.
  The space count display is updated (at the end of recursion).

*/


void Minefield::Clear(Tile* t)
{
  static int recur_count = 0;

  ++recur_count;

  if (Game == CLEAR)
  {
    static struct OpenQueue
    {
      Tile* t;
      OpenQueue* next;
      OpenQueue() : next(NULL) {}
    } *head = new OpenQueue, *tail = head, *temp;

    for (;;)
    {
      around (t, loop)
      {
	if (t->surround[loop]->status == UNOPENED)
        {
	  Open(t->surround[loop]); if (isFinished) break;
	  if (t->surround[loop]->value == 0 ||
	      (isAutoOpen && t->surround[loop]->value ==
					     Count(t->surround[loop], MARKED)))
	  {
	    tail->t = t->surround[loop];
	    if (tail->next == NULL) tail->next = new OpenQueue;
	    tail = tail->next;
          }
        }
      }
      if (isFinished) tail = head;
      if (head == tail) break;
      t = head->t;
      temp = head;			// Make the tail point to this one
      head = head->next;		// and this one point to the one
      temp->next = tail->next;		// after the tail.
      tail->next = temp;
    }
  }
  else                                  // The cross game
  {
    around (t, loop)
    {
      if (t->surround[loop]->status == UNOPENED)
      {
	Open(t->surround[loop]); if (isFinished) break;
	if (t->surround[loop]->value == 0 ||
	    (isAutoOpen && t->surround[loop]->value ==
					     Count(t->surround[loop], MARKED)))
	{
	  Clear(t->surround[loop]); if (isFinished) break;
	}
      }
    }
  }
  if (--recur_count == 0) Board->DisplayProgress(mines, spaces, UPDATE_SPACES);
}


/* ----------------------------   ClearAround   -----------------------------

  If the number of marked mines around t equals the value of t then
  open all the surrounding tiles. If it doesn't then highlight the unopened
  tiles (provided doHiLite is true).

*/

void Minefield::ClearAround(Tile* t, bool doHiLite)
{
  if (t->status != OPENED) return;
  int m = Count(t, MARKED);
  if (t->value == m) Clear(t);
  else if (doHiLite)
  {
    HighLight(t, (t->value > m ? MINE : OPENED));
    ht = t;
    isHiLite = true;
  }
}


/* -----------------------------   HighLight   ------------------------------

  Highlight the tiles that could have a mine for clear-around. The highlight
  is removed when the mouse button is released.
  If type is MINE, unopened tiles are drawn with a mine; if type is OPENED,
  marked tiles are drawn opened. If type is UNOPENED, the highlight is removed.

  There currently cannot be more than one highlight.

*/

void Minefield::HighLight(Tile* t, int type)
{
       if (type == MINE)   hs = UNOPENED;
  else if (type == OPENED) hs = MARKED;

  around (t, loop)
  {
    if (t->surround[loop]->status == hs)
      DisplayTile(t->surround[loop], type);
  }
}

void Minefield::UnHighLight()
{
  if (isHiLite) HighLight(ht, hs);
  isHiLite = false;
}


/* --------------------------------   Mark   --------------------------------

  Mark an unopened tile with a flag; change a marked tile to an unsure
  tile (if the Marks options is set) or remove a mark. Update the flag count
  and test for game over (Finish & ALL_MARKED).
  If automatic opening is on then test the tiles around the one being marked.
  The tile is assumed to be unopened (ie. not OPENED).

  This function becomes recursive when both automatics are set (Open() calls
  Mark() calls ClearAround() calls Clear() calls Open() calls Mark()...).
  Don't update the flag count until recursion is over.

*/

void Minefield::Mark(Tile* t)
{
  static int recur_count = 0;

  switch (t->status)
  {
    case UNOPENED: --mines; mynes -= (t->value == MINE);
		   t->status = MARKED;
		   break;

    case MARKED:   ++mines; mynes += (t->value == MINE);
		   if (isMarks)
		   {
		     t->status = UNSURE;
		     break;
		   }
		   // else fall through
    case UNSURE:   t->status = UNOPENED;
  }

  if (isSound && MarkSnd) PlaySampleFVP(mark_snd, mark_snd->frequency,
					mark_snd->volume, PanLevel(t->loc));
  DisplayTile(t, t->status);
  if (mynes == 0) if (GameOver()) return;

  ++recur_count;
  if (isAutoOpen)
  {
    around (t, loop)
    {
      ClearAround(t->surround[loop], false);
      if (isFinished) break;
    }
  }
  if (--recur_count == 0) Board->DisplayProgress(mines, spaces, UPDATE_MINES);
}


/* ------------------------------   GameOver   ------------------------------

  The game is over when a mine is opened (dead == true); when all the non-mine
  tiles have been opened (Finish & ALL_OPENED), or all the mine tiles have been
  marked (Finish & ALL_MARKED), or both; or the field has been crossed.
  If the game was lost then all the mines and incorrect marks are displayed.
  If the game was won then all the mines are marked and all the tiles are
  opened (clear), or all opened tiles are displayed as blanks (cross). In the
  cross game, the counts show the number of mines flagged and tiles opened.

*/

bool Minefield::GameOver(bool dead)
{
  if (!dead)
  {
    if (Game == CLEAR)
    {
      if (!((spaces == 0 && (Finish & ALL_OPENED)) ||
	    (mynes  == 0 && (Finish & ALL_MARKED)) ||
	    (spaces == 0 && mynes == 0 && Finish == 0))) return false;
    }
    else
    {
      if (CrossEnd()->status != OPENED) return false;
    }
  }
  int score = Board->GameOver(dead);
  isFinished = true;

  if (dead)
  {
    if (isSound && DieSnd) PlaySample(die_snd);
    foralltiles
    {
      if (tiles[tile]->value == MINE)
      {
	if (tiles[tile]->status == UNOPENED || tiles[tile]->status == UNSURE)
	  DisplayTile(tiles[tile], MINE);
      }
      else if (tiles[tile]->status == MARKED)
      {
	DisplayTile(tiles[tile], WRONGMINE);
	Board->DisplayProgress(++mines, spaces, UPDATE_MINES);
      }
    }
  }
  else
  {
    if (isSound && WinSnd) PlaySample(win_snd);
    if (Game == CLEAR)
    {
      spaces = mines = 0;
    }
    else
    {
      spaces = (area(layout.dim) - layout.mines) - spaces;
      mines  = layout.mines - mines;
    }
    Board->DisplayProgress(mines, spaces, UPDATE_BOTH);
    foralltiles
    {
      if (Game == CLEAR)
      {
	if (tiles[tile]->status == UNOPENED || tiles[tile]->status == UNSURE)
	  DisplayTile(tiles[tile],
		   (tiles[tile]->value == MINE) ? MARKED : tiles[tile]->value);
      }
      else if (tiles[tile]->status == OPENED) DisplayTile(tiles[tile], OPENED);
    }
    times.Check(level, score, spaces, mines);
  }
  return true;
}


/* -------------------------------   Pause   --------------------------------

  Blank out the minefield.

*/

void Minefield::Pause()
{
  foralltiles DisplayTile(tiles[tile], OPENED, false);
}


/* ------------------------------   PanLevel   ------------------------------

  Return the panning associated with the tile being opened or marked, in
  relation to the minefield, not the screen.

*/

int Minefield::PanLevel(cCoord& c)
{
  return (65536 * ((SwapStereo) ? c.x : (pan_wid - c.x)) / pan_wid);
}


/* ------------------------------   ReadIni   -------------------------------

  Read the custom and random games, and level, from the initialisation file,
  verify that those games are valid, and load the times.
  Note that the custom and random layouts have to be initialised before the
  call to SetLevel().

*/

void Minefield::ReadIni(IniFile& ini, const char* section, const char* file)
{
  Custom.dim.x = ini.GetInt(section, "Custom.Width");
  Custom.dim.y = ini.GetInt(section, "Custom.Height");
  Custom.mines = ini.GetInt(section, "Custom.Mines");
  Validate(Custom);

  Random.dim.x = ini.GetInt(section, "Random.Width");
  Random.dim.y = ini.GetInt(section, "Random.Height");
  Random.mines = ini.GetInt(section, "Random.Mines");
  Validate(Random);

  SetLevel(ini.GetInt(section, "Level", BEGINNER));

  field = this;
  times.Load(file, CustomPopup);
}


/* ------------------------------   WriteIni   ------------------------------

  Write the level and custom and random games to the initialisation file, and
  save the times.

*/

void Minefield::WriteIni(IniFile& ini, const char* section, const char* file)
{
  ini.WriteInt(section, "Level",         level);

  ini.WriteInt(section, "Custom.Width",  Custom.dim.x);
  ini.WriteInt(section, "Custom.Height", Custom.dim.y);
  ini.WriteInt(section, "Custom.Mines",  Custom.mines);

  ini.WriteInt(section, "Random.Width",  Random.dim.x);
  ini.WriteInt(section, "Random.Height", Random.dim.y);
  ini.WriteInt(section, "Random.Mines",  Random.mines);

  times.Save(file);
}
