/*
  Program:      Mine Mayhem
  File: 	minefld.h
  Date: 	July, 1996 to February,   1997, version 1.00.
		26 October to 3 December,	version 2.00.
		8 December to 4 January,  2001, version 2.10.
		15 January,			version 2.11.

  Definitions for the minefield.

  Freeware, Copyright 2001 Jason Hood.

  You are free to use this code, or a portion thereof, as long as an
  appropriate acknowledgement is made.

  v2.00: each tile knows it's displayed location and surrounding tiles;
	 lots of changes due to the board -> minefield conversion.
  v2.10: modified handling of panning.
*/

#ifndef _minefield_h
#define _minefield_h

#include "layout.h"
#include "custom.h"
#include "tiles.h"
#include "inifile.h"


// Are we clearing or crossing the field?
extern int Game;
#define CLEAR		0
#define CROSS           1

#include "times.h"              // Needs Game defined

// Is it a square-tiled, hexagon-tiled, or hexagonal minefield?
extern int GameStyle;
#define SQUARES   0
#define HEXAGONS  1
#define HEXAGONAL 2
#define NoGameStyles 3


// The levels - 2 to 10 are the appropriate random games.
enum { BEGINNER = 11, INTERMEDIATE, EXPERT, CUSTOM };


// A tile on the minefield
struct Tile
{
  int	 value, 			// Number of mines around the tile
					//  or MINE for a mine.
	 status;			// UNOPENED, OPENED, MARKED, UNSURE.

  Coord  loc;				// Displayed location.

  Tile** surround;			// Surrounding tiles.
  int	 count; 			// Number of above.

  ~Tile() { delete[] surround; }
};


class Minefield
{
public:
	   Minefield();
  virtual ~Minefield();


  virtual int	area(cCoord& dim) = 0;

  virtual bool	canWrap() = 0;

	  void	SetLevel(int Level, int from_ini = true);
	  void	NewGame(int Level = 0);
  virtual void	AdjustGame() = 0;	// Changing from CLEAR to CROSS
  virtual void	AdjustWrap() = 0;	// Changing the Wrap-around option

	  bool	GameinProgress();

	  void	LeftButton(int x, int y);
	  void	RightButton(int x, int y);
	  void	MiddleButton(int x, int y);

	  void	UnHighLight();		// Remove the highlight
  virtual bool	Convert(int x, int y, Tile*& t) = 0;	// Mouse co-ord to tile

  virtual Coord GetBoardDimensions(cCoord& dim) 			 = 0;
  virtual void	Outline(ViewBuffer* buf, int x1, int y1, int x2, int y2) = 0;
	  void	Pause();


  // By rights these should be protected, with public functions to access them,
  // but I couldn't be bothered. (Why can't we have variables read-only outside
  // the class?)

  Layout Beginner,			// These three should be obvious
	 Intermediate,
	 Expert,
	 Custom,			// Initially from the initialisation
	 Random,			//  file (hmmm... :) )
	 layout;			// The layout being played

  // The minimum and maximum dimensions the minefield can have.
  int MinWidth, MaxWidth, MinHeight, MaxHeight;

  CustomPopupRes CustomPopup;		// The stored custom games
  Times times;				// All the times for all the levels

  int level;				// Current level

protected:
	  void	ReadIni( IniFile& ini, const char* section, const char* file);
	  void	WriteIni(IniFile& ini, const char* section, const char* file);
  virtual void	Validate(Layout& layout) = 0;

	  void	CreateMinefield(int width);

  virtual Tile* CrossStart() = 0;
  virtual Tile* CrossEnd()   = 0;

  virtual void	EveryTile(bool create = false) = 0;
  virtual void	Surround(int x, int y)	       = 0;
  virtual void	Validate(int x, int y)	       = 0;

  virtual void	RandomLayout(int diff) = 0;

	  bool	CrossValid(Tile* t);		// Co-ord valid for cross game?

	  void	PlaceMine();
	  bool	MoveMine(Tile* t);
	  bool	NearestSpace(Tile*& t); 	// Find the closest empty tile

	  void	Open(Tile* t);
	  void	Clear(Tile* t);
	  void	ClearAround(Tile* t, bool doHiLite = true);
	  void	HighLight(Tile* t, int type);
	  void	Mark(Tile* t);
	  bool	GameOver(bool dead = false);

	  int	Count(Tile* t, int what);

  virtual void	DisplayTile(Tile* t, int which, bool refresh = true) = 0;

	  int	PanLevel(cCoord& c);	// Panning for sound


  int	 max_width, max_height; 	// Available screen dimensions
  int	 rows_allocated;		// For deleting the grid

  Tile** grid;				// The minefield
  Tile** tiles; 			// Pointers to every tile
  int	 tile_count;			// Number of above

  bool	 isStarted,			// The game has begun
	 isFinished;			// The game is over

  int	 spaces,			// Spaces cleared
	 mines, 			// Mines flagged
	 mynes; 			// Mines found

  Tile*  ht;				// The highlight centre position
  int	 hs;				// The status to be highlighted
  bool	 isHiLite;			// Is it on?

  Tile*  tile;				// Current tile being surrounded

  int	 pan_wid;			// Value to determine panning level
};

// These are meant for private use.
#define around(t, loop) \
for (int loop = 0; loop < t->count; ++loop)

#define foralltiles \
for (int tile = 0; tile < tile_count; ++tile)


class SquareField : public Minefield
{
public:
  SquareField(bool create = true);
  void	ReadIni( IniFile& ini);
  void	WriteIni(IniFile& ini);

  int	area(cCoord& dim);

  bool	canWrap() { return true; }

  void	AdjustGame();
  void	AdjustWrap();

  Coord GetBoardDimensions(cCoord& field);
  void	Outline(ViewBuffer* buf, int x1, int y1, int x2, int y2);

protected:
  Tile* CrossStart();
  Tile* CrossEnd();

  void	EveryTile(bool create = false);
  void	Surround(int x, int y);
  void	Validate(int x, int y);
  void	AdjustSurround();

  void	RandomLayout(int diff);

  bool	Convert(int x, int y, Tile*& t);

  void	DisplayTile(Tile* t, int which, bool refresh = true);

  int	PanLevel(cCoord& c);

  void	Validate(Layout& layout);

  Coord olddim; 			// Used by AdjustSurround()
};


class RectHexField : public SquareField
{
public:
  RectHexField();
  void	ReadIni( IniFile& ini);
  void	WriteIni(IniFile& ini);

  bool	canWrap();

  Coord GetBoardDimensions(cCoord& field);
  void	Outline(ViewBuffer* buf, int x1, int y1, int x2, int y2);

protected:

  void	EveryTile(bool create = false);
  void	Surround(int x, int y);
  void	AdjustSurround();

  bool	Convert(int x, int y, Tile*& t);

  void	DisplayTile(Tile* t, int which, bool refresh = true);

  int	PanLevel(cCoord& c);
};


class HexagonField : public Minefield
{
public:
  HexagonField();
  void	ReadIni( IniFile& ini);
  void	WriteIni(IniFile& ini);

  int	area(cCoord& dim);

  bool	canWrap() { return false; }

  void	AdjustGame();
  void	AdjustWrap();

  Coord GetBoardDimensions(cCoord& field);
  void	Outline(ViewBuffer* buf, int x1, int y1, int x2, int y2);

  int	GetMaxWidth(int height);

protected:
  Tile* CrossStart();
  Tile* CrossEnd();

  void	EveryTile(bool create = false);
  void	Surround(int x, int y);
  void	Validate(int x, int y);

  void	RandomLayout(int diff);

  bool	Convert(int x, int y, Tile*& t);

  void	DisplayTile(Tile* t, int which, bool refresh = true);

  int	PanLevel(cCoord& c);

  void	Validate(Layout& layout);

  int totalcols;			// The hexagon's maximum middle width
};


extern Minefield*    field;
extern SquareField*  squarefield;
extern RectHexField* recthexfield;
extern HexagonField* hexagonfield;

#endif
